unit HtmlHighlight;
{  Electro-Concept Mauricie, 1998-1999-2000
  Version 5.2
  June 22, 2000 release

  New since April 16, 2000:
     Added Perl script highlighting: sections inside <? ... ?> will be highlighted according
     to ScriptContent, and the scope for this section will be set to PerlScriptScope

     Added property XMLSyntax: if True, every tag will be highlighted like html tags, i.e. it
     does not need to begin with a valid html keyword.  There will be no Unknown tag anymore.

  A component to dynamically highlight html language elements in a TPlusMemo.

  Place a THTMLHighlighter on your form (it will appear on the ECM page as default), set your
  TPlusMemo.Highlighter property to point to it, et voil, html tags are turned colored,
  underlined, bolded, as you want.  Various parts of html tags are highlighted independantly,
  for example the bracket pair, the attribute values and strings can have their own highlighting,
  different from the html keyword.  Html comments and script sections are also highlighted
  differently.

  Properties DefaultText and ScriptTag are there only for the purpose of compatibility with earlier version.

  Note: this component is based on TCustomExtHighlighter, a class defined in file ExtHilit.pas.
  It is possible that you have already this file on your system, because it is also used for the
  TExtHighlighter and TOOPHighlighter accessories.  If so, please keep only the most recent one.

  You can control the highlighting attributes for different parts of html tags from within the Object
  Inspector. This effect is best appreciated if some html text is loaded in the TPlusMemo (via its
  Lines property), and watching how the highlighting changes when changing various properties in the
  Object Inspector }

{ License terms:
----------------
  This component is freeware. You may copy those files AS LONG AS YOU COPY ALL OF THEM and don't charge
  any kind of fee for it. If you want to change the source code in order to improve the components features,
  performance, etc. please send us the new source code so that we can have a look at it. The changed source
  code should contain descriptions what you have changed, and of course your name. The only thing you MAY NOT
  CHANGE is the ORIGINAL COPYRIGHT INFORMATION.

  Disclaimer
----------
   The author is NOT liable for any DAMAGES resulting from the use and misuse of the component, especially
   he is NOT liable for DAMAGES that were caused BY ANY VERSION WHICH HAS NOT BEEN PROGRAMMED BY THE ORIGINAL
   AUTHOR HIMSELF.

Contact us
----------
  If you have questions, suggestions or comments, feel free to send an E-Mail to
  Electro-Concept Mauricie     ecm.@tr.cgocable.ca

Thank you
  }

{$A+}  { this unit requires word alignment of data }
{$B-}  { not complete boolean evaluation }

{$IFDEF WIN32}
  {$H+}  { long strings }
  {$J+}  { writeable typed constants }
{$ENDIF}
{$IFNDEF DEBUG}
  {$R-}
  {$Q-}
{$ENDIF}

interface

uses Classes, Graphics, Controls, Plusmemo, ExtHilit;

const
  HTMLTagScope = 1;
  ScriptScope  = 2;
  HTMLCommentScope= 3;
  HTMLStartScope  = 4;
  HtmlKeywordScope= 5;
  HtmlUnknownScope= 6;
  HtmlStopScope   = 7;
  HtmlAttributeScope = 8;
  HtmlAttributeValueScope = 9;
  SpecialCharScope = 10;
  PerlScriptScope = 11;


type THtmlHighlighter = class(TCustomExtHighlighter)
       private
         fBracket,
         fDefaultText,
         fHtmlTag,
         fHtmlComment,
         fHtmlAttribute,
         fHtmlAttributeValue,
         fScriptTag,
         fScriptContent,
         fUnknownTag,
         fHtmlKeyword,
         fSpecialChars: THighlightInfo;

         fXMLSyntax: Boolean;

         procedure HighlightChange(Sender: TObject);
         procedure setBracket(const Value: THighlightInfo);
         procedure setHtmlTag(ht: THighlightInfo);
         procedure setHtmlComment(hc: THighlightInfo);
         procedure setHtmlAttribute(ha: THighlightInfo);
         procedure setHtmlAttributeValue(ha: THighlightInfo);
         procedure setScriptTag(st: THighlightInfo);
         procedure setScriptContent(sc: THighlightInfo);
         procedure setSpecialChars(sc: THighlightInfo);
         procedure setUnknownTag(ut: THighlightInfo);
         procedure setHtmlKeyword(const Value: THighlightInfo);

       protected
         procedure Loaded; override;
         function  FindStart(Start, Stop: TPlusNavigator; BaseIndex: Integer): Boolean; override;
         function  FindStop(Start, Stop: TPlusNavigator; BaseIndex: Integer): Boolean; override;

       public
         constructor Create(AOwner: TComponent); override;
         destructor Destroy; override;

       published
         property Bracket      : THighlightInfo read fBracket       write setBracket;
         property DefaultText  : THighlightInfo read fDefaultText   write fDefaultText;
                                         { only for compatibility with pre-4.1 version }
         property HtmlKeyword  : THighlightInfo read fHtmlKeyword   write setHtmlKeyword;
         property HtmlTag      : THighlightInfo read fHtmlTag       write setHtmlTag;
         property HtmlComment  : THighlightInfo read fHtmlComment   write setHtmlComment;
         property HtmlAttribute: THighlightInfo read fHtmlAttribute write setHtmlAttribute;
         property HtmlAttributeValue: THighlightInfo read fHtmlAttributeValue write setHtmlAttributeValue;
         property ScriptTag    : THighlightInfo read fScriptTag     write setScriptTag;
                                         { only for compatibility with pre-4.1 version }
         property ScriptContent: THighlightInfo read fScriptContent write setScriptContent;
         property SpecialChars : THighlightInfo read fSpecialChars  write setSpecialChars;
         property UnknownTag   : THighlightInfo read fUnknownTag    write setUnknownTag;
         property SubHighlighter;  { promoted from TCustomExtHighlighter to allow nested highlighting, ex:
                                     url highlighter within comment sections }

         property XMLSyntax: Boolean read fXMLSyntax write fXMLSyntax;
       end;


procedure Register;

implementation

uses SysUtils;

type kstring = {$IFDEF WIN32} string; {$ELSE} string[20] {$ENDIF}

const
  DefaultTextColor = clWindowText;
  DefaultBackColor = -1;

  HTMLTagColor      = clGreen;
  HTMLTagBackground = -1;
  HTMLCommentColor  = clOlive;
  HTMLKeywordColor  = clNavy;
  HTMLAttributeColor = clGreen;
  HTMLAttributeValueColor = clGray;

  ScriptTagBackColor   = clWindow;
  ScriptTagForeColor   = clPurple;
  ScriptTextColor      = clWindowText;
  UnknownTagColor      = clRed;


  ScriptBackColor   = -1;
  ScriptForeColor   = clNavy;
  BracketColor      = clBlue;
  SpecialCharColor  = clFuchsia;

  HtmlKeywordCount  = 106;

type HtmlElementRecord = record Element: kstring; Attributes, Complement: Boolean end;
     HtmlElementArrays = array[0..HtmlKeywordCount-1] of HtmlElementRecord;

const HtmlList: HtmlElementArrays =

    ((Element: 'SCRIPT';       Attributes: True;  Complement: False),    { this one put first because of special treatment }
     (Element: 'A';            Attributes: True;  Complement: True ),
     (Element: 'ADDRESS';      Attributes: True;  Complement: True ),
     (Element: 'ACRONYM';      Attributes: True;  Complement: True ),
     (Element: 'APPLET';       Attributes: True;  Complement: True ),
     (Element: 'AREA';         Attributes: True;  Complement: True ),
     (Element: 'B';            Attributes: False; Complement: True ),
     (Element: 'BASE';         Attributes: True;  Complement: False),
     (Element: 'BASEFONT';     Attributes: True;  Complement: False),
     (Element: 'BGSOUND';      Attributes: True;  Complement: True ),
     (Element: 'BIG';          Attributes: True;  Complement: True ),
     (Element: 'BLINK';        Attributes: True;  Complement: True ),
     (Element: 'BLOCKQUOTE';   Attributes: True;  Complement: True ),
     (Element: 'BODY';         Attributes: True;  Complement: True ),
     (Element: 'BR';           Attributes: False; Complement: False),
     (Element:  'BUTTON';      Attributes: True;  Complement: True ),
     (Element:  'CAPTION';     Attributes: True;  Complement: True ),
     (Element:  'CITE';        Attributes: True;  Complement: True ),
     (Element:  'CENTER';      Attributes: True;  Complement: True ),
     (Element:  'CODE';        Attributes: True;  Complement: True ),
     (Element:  'COL';         Attributes: True;  Complement: True ),
     (Element:  'COLGROUP';    Attributes: True;  Complement: True ),
     (Element:  'COMMENT';     Attributes: True;  Complement: True ),
     (Element:  'DD';          Attributes: True;  Complement: True ),
     (Element:  'DEL';         Attributes: True;  Complement: True ),
     (Element:  'DFN';         Attributes: True;  Complement: True ),
     (Element:  'DIR';         Attributes: True;  Complement: True ),
     (Element:  'DIV';         Attributes: True;  Complement: True ),
     (Element:  'DL';          Attributes: True;  Complement: True ),
     (Element:  'DT';          Attributes: True;  Complement: True ),
     (Element:  'EM';          Attributes: True;  Complement: True ),
     (Element:  'EMBED';       Attributes: True;  Complement: True ),
     (Element:  'FIELDSET';    Attributes: True;  Complement: True ),
     (Element:  'FORM';        Attributes: True;  Complement: True ),
     (Element:  'FONT';        Attributes: True;  Complement: True ),
     (Element: 'FRAME';        Attributes: True;  Complement: False),
     (Element:  'FRAMESET';    Attributes: True;  Complement: True ),
     (Element:  'H1';          Attributes: True;  Complement: True ),
     (Element:  'H2';          Attributes: True;  Complement: True ),
     (Element:  'H3';          Attributes: True;  Complement: True ),
     (Element:  'H4';          Attributes: True;  Complement: True ),
     (Element:  'H5';          Attributes: True;  Complement: True ),
     (Element:  'H6';          Attributes: True;  Complement: True ),
     (Element:  'HEAD';        Attributes: False; Complement: True ),
     (Element:  'HR';          Attributes: True;  Complement: True ),
     (Element:  'HTML';        Attributes: False; Complement: True ),
     (Element:  'I';           Attributes: False; Complement: True ),
     (Element:  'IFRAME';      Attributes: True;  Complement: True ),
     (Element:  'ILAYER';      Attributes: True;  Complement: True ),
     (Element:  'IMG';         Attributes: True;  Complement: False),
     (Element:  'INPUT';       Attributes: True;  Complement: False),
     (Element:  'INS';         Attributes: True;  Complement: True ),
     (Element:  'ISINDEX';     Attributes: True;  Complement: True ),
     (Element:  'KBD';         Attributes: True;  Complement: True ),
     (Element:  'KEYGEN';      Attributes: True;  Complement: True ),
     (Element:  'LABEL';       Attributes: True;  Complement: True ),
     (Element:  'LAYER';       Attributes: True;  Complement: True ),
     (Element:  'LEGEND';      Attributes: True;  Complement: True ),
     (Element:  'LI';          Attributes: True;  Complement: True ),
     (Element:  'LINK';        Attributes: True;  Complement: True ),
     (Element:  'LISTING';     Attributes: True;  Complement: True ),
     (Element:  'MAP';         Attributes: True;  Complement: True ),
     (Element:  'MARQUEE';     Attributes: True;  Complement: True ),
     (Element:  'MENU';        Attributes: True;  Complement: True ),
     (Element:  'META';        Attributes: True;  Complement: False),
     (Element:  'MULTICOL';    Attributes: True;  Complement: True ),
     (Element:  'NEXTID';      Attributes: True;  Complement: True ),
     (Element:  'NOBR';        Attributes: True;  Complement: True ),
     (Element:  'NOEMBED' ;    Attributes: True;  Complement: True ),
     (Element:  'NOFRAMES';    Attributes: True;  Complement: True ),
     (Element:  'NOLAYER';     Attributes: True;  Complement: True ),
     (Element:  'NOSCRIPT';    Attributes: True;  Complement: True ),
     (Element:  'OBJECT';      Attributes: True;  Complement: True ),
     (Element:  'OL';          Attributes: True;  Complement: True ),
     (Element:  'OPTION';      Attributes: True;  Complement: True ),
     (Element:  'P';           Attributes: True;  Complement: True ),
     (Element:  'PARAM';       Attributes: True;  Complement: True ),
     (Element:  'PLAINTEXT';   Attributes: True;  Complement: True ),
     (Element:  'Q';           Attributes: True;  Complement: True ),
     (Element:  'S';           Attributes: False; Complement: True ),
     (Element:  'SAMP';        Attributes: True;  Complement: True ),
     (Element:  'SELECT';      Attributes: True;  Complement: True ),
     (Element:  'SOUND';       Attributes: True;  Complement: True ),
     (Element:  'SMALL';       Attributes: True;  Complement: True ),
     (Element:  'SPACER';      Attributes: True;  Complement: True ),
     (Element:  'SPAN';        Attributes: True;  Complement: True ),
     (Element:  'STRIKE';      Attributes: True;  Complement: True ),
     (Element:  'STRONG';      Attributes: True;  Complement: True ),
     (Element:  'STYLE';       Attributes: True;  Complement: True ),
     (Element:  'SUB';         Attributes: True;  Complement: True ),
     (Element:  'SUP';         Attributes: True;  Complement: True ),
     (Element:  'TABLE';       Attributes: True;  Complement: True ),
     (Element:  'TBODY';       Attributes: True;  Complement: True ),
     (Element:  'TD';          Attributes: True;  Complement: True ),
     (Element:  'TEXTAREA';    Attributes: True;  Complement: True ),
     (Element:  'TFOOT';       Attributes: True;  Complement: True ),
     (Element:  'TH';          Attributes: True;  Complement: True ),
     (Element:  'THEAD';       Attributes: True;  Complement: True ),
     (Element:  'TITLE';       Attributes: False; Complement: True ),
     (Element:  'TR';          Attributes: True;  Complement: True ),
     (Element:  'TT';          Attributes: True;  Complement: True ),
     (Element:  'U';           Attributes: False; Complement: True ),
     (Element:  'UL';          Attributes: True;  Complement: True ),
     (Element:  'VAR';         Attributes: True;  Complement: True ),
     (Element:  'WBR';         Attributes: True;  Complement: True ),
     (Element:  '!DOCTYPE';    Attributes: True;  Complement: False ));

  HtmlCommentIndex = HtmlKeywordCount;
  HtmlUnknownIndex = HtmlCommentIndex + 1;
  PcScriptIndex    = HtmlUnknownIndex + 1;    { % script index }
  SpecialCharIndex = PcScriptIndex + 1;
  PerlScriptIndex  = SpecialCharIndex + 1;
  HtmlScriptIndex  = 0;
  HtmlIndexCount   = PerlScriptIndex + 1;

procedure Register;
begin
  RegisterComponents('ECM', [THTMLHighlighter])
end;

function getlevel(const dyn: DynInfoRec): SmallInt;
begin
  if dyn.DynStyle and $c0 <>$c0 then Result:= -1
                                else Result:= dyn.Level
end;

function HStrScan(t: PChar; c: Char): PChar;  { better than the one supplied in SysUtils for long PChars }
begin
  while (t^<>#0) and (t^<>c) do Inc(t);
  if t^=#0 then Result:= nil else Result:= t
end;

procedure FillDyn(h: THighlightInfo; nav: TPlusNavigator; Nested: Boolean; Ctx, KIndex: Integer; StartLen, StopLen: PlusCardinal);
var dinfo: DynInfoRec; oldpos: Longint;
begin
  with dinfo do
    begin
      Level:= getlevel(nav.DynAttr);
      if (Level>=0) and Nested then
        with nav do
          begin
            oldpos:= Pos;
            if BackToDyn(0) then
              begin
                Par^.DynBuffer^[DynNb].StopKLen:= 0;
                Pos:= oldpos;
                if Nested then RightOfDyn
              end
          end;

      if Nested or (Level<0) then Inc(Level);
      DynStyle:= Byte(h.Style) or $c0;
      DynOffset:= nav.ParOffset;
      KeyIndex:= nav.DynAttr.KeyIndex;
      if h.AltFont then DynStyle:= DynStyle or (1 shl Ord(fsAltFont));
      KeyIndex[Level]:= -32768+ KIndex;
      Cursor:= crDefault;
      Backgnd:= h.Background;
      Foregnd:= h.Foreground;
      Context:= Ctx;
      StartKlen:= StartLen;
      StopKLen:= StopLen;
      with nav do
        if (fDynNb<fPar^.DynCount) and (fPar^.DynBuffer^[fDynNb].DynOffset=fOffset) then
                       fPar^.DynBuffer^[fDynNb]:= dinfo
        else AddDyn(dinfo);
    end
end;

constructor THtmlHighlighter.Create(AOwner: TComponent);
  procedure CreateHighlight(var h: THighlightInfo; style: TExtFontStyles; bk, fr: TColor);
  begin
    h:= THighlightInfo.Create;
    h.AltFont:=  Byte(style) and (1 shl Ord(fsAltFont)) <> 0;
    h.Style:= TFontStyles(Byte(Byte(style) and (not (1 shl Ord(fsAltFont)))));// - [TExtFontStyle(fsAltFont)];
    h.Background:= bk;
    h.Foreground:= fr;
    h.OnChange:= HighlightChange
  end;

var afstyle: TExtFontStyles;

begin
  inherited Create(AOwner);
  Byte(afstyle):= 1 shl Ord(fsAltFont);
  CreateHighlight(fDefaultText,[], DefaultBackColor, DefaultTextColor);
  CreateHighlight(fHtmlKeyword,[fsBold],DefaultBackColor,HtmlKeywordColor);
  CreateHighlight(fHtmlTag, [fsItalic], DefaultBackColor, HTMLTagColor);
  CreateHighlight(fHtmlComment, [fsItalic], DefaultBackColor, HTMLCommentColor);
  CreateHighlight(fHtmlAttribute, [], DefaultBackColor, HTMLAttributeColor);
  CreateHighlight(fHtmlAttributeValue, [], DefaultBackColor, HTMLAttributeValueColor);
  CreateHighlight(fScriptTag, [fsBold], ScriptTagBackColor, ScriptTagForeColor);
  CreateHighlight(fScriptContent, afstyle, DefaultBackColor, ScriptTextColor);
  CreateHighlight(fUnknownTag, [], DefaultBackColor, UnknownTagColor);
  CreateHighlight(fBracket, [fsBold], DefaultBackColor, BracketColor);
  CreateHighlight(fSpecialChars, [], DefaultBackColor, SpecialCharColor);
end;


destructor THtmlHighlighter.Destroy;
begin
  fHtmlKeyword.Destroy;
  fHtmlTag.Destroy;
  fHtmlComment.Destroy;
  fHtmlAttribute.Destroy;
  fHtmlAttributeValue.Destroy;
  fScriptTag.Destroy;
  fDefaultText.Destroy;
  fScriptContent.Destroy;
  fUnknownTag.Destroy;
  fBracket.Destroy;
  fSpecialChars.Destroy;
  inherited Destroy
end;

procedure THtmlHighlighter.HighlightChange(Sender: TObject);
    { event handler for internal THighlightInfo modifications }
begin
  if (csDesigning in ComponentState) and (not (csLoading in ComponentState)) then ReApplyKeys
end;   { THtmlHighlighter.HighlightChange }

procedure THtmlHighlighter.Loaded;
begin
  inherited Loaded;
  ReApplyKeys
end;


procedure THtmlHighlighter.setBracket(const Value: THighlightInfo);
begin
  fBracket.Assign(Value);
end;

procedure THtmlHighlighter.setHtmlKeyword(const Value: THighlightInfo);
begin
  fHtmlKeyword.Assign(Value);
end;

procedure THtmlHighlighter.setHtmlTag(ht: THighlightInfo);
begin
  fHtmlTag.Assign(ht)
end;

procedure THtmlHighlighter.setHtmlComment(hc: THighlightInfo);
begin
  fHtmlComment.Assign(hc)
end;

procedure THtmlHighlighter.setHtmlAttribute(ha: THighlightInfo);
begin
  fHtmlAttribute.Assign(ha)
end;

procedure THtmlHighlighter.setHtmlAttributeValue(ha: THighlightInfo);
begin
  fHtmlAttributeValue.Assign(ha)
end;

procedure THtmlHighlighter.setScriptTag(st: THighlightInfo);
begin
  fScriptTag.Assign(st)
end;

procedure THtmlHighlighter.setScriptContent(sc: THighlightInfo);
begin
  fScriptContent.Assign(sc)
end;

procedure THtmlHighlighter.setSpecialChars(sc: THighlightInfo);
begin
  fSpecialChars.Assign(sc)
end;

procedure THtmlHighlighter.setUnknownTag(ut: THighlightInfo);
begin
  fUnknownTag.Assign(ut)
end;

function THtmlHighlighter.FindStop(Start, Stop: TPlusNavigator; BaseIndex: Integer): Boolean; { they must be in the same par. }
var searchlen  : PlusCardinal;
    backstart, backend: PlusCardinal;
    dinfo      : DynInfoRec;
    t, keyfound: PChar;
    scar       : Char;
    curscope,
    curlevel, curkey   : Integer;
    stopendpos : Longint;

const EndComment = '-->'#0;
      EndScript  = '%>'#0;
      EndPerlScript = '?>'#0;
      EndHtmlScript = '</SCRIPT>'#0;

begin
  Result:= False;

  Start.RightOfDyn;
  curlevel:= getlevel(Start.pDynAttr^);
  with Start.pDynAttr^ do
    if DynStyle and $80 = 0 then
      begin
        curscope:= 0;
        curkey:= High(curkey)
      end
    else
      begin
        curscope:= Context;
        if curlevel>=0 then curkey:= KeyIndex[curlevel]
                       else curkey:= High(curkey);
        if curkey<0 then curkey:= curkey + 32768
                    else curkey:= High(curkey)
      end;

  if (curkey-BaseIndex>=HtmlIndexCount) and (SubHighlighter<>nil) then
    begin
      Result:= inherited FindStop(Start, Stop, HtmlIndexCount+BaseIndex);
      Exit
    end;

  case curscope of
    HtmlTagScope, HtmlUnknownScope:
      begin
        t:= Start.fPar^.Text + Start.fOffset;
        if t=nil then Exit;
        searchlen:= Stop.ParOffset - Start.fOffset;
        scar:= t[searchlen];
        t[searchlen]:= #0;
        keyfound:= HStrScan(t, '>');
        if keyfound<>nil then
          begin
            Result:= True;
            Stop.Assign(Start);
            Stop.Pos:= Start.Pos + (keyfound - t);
            FillDyn(Bracket, Stop, False, HtmlStopScope, BaseIndex, 1, 2);
            Stop.Pos:= Stop.Pos + 1;
            if (curkey=HtmlScriptIndex+BaseIndex) then
              FillDyn(ScriptContent, Stop, True, ScriptScope, BaseIndex + HtmlScriptIndex, 0, 10)
            else
              begin
                dinfo.DynStyle:= 0;
                Stop.AddDyn(dinfo)
              end
          end;
        t[searchlen]:= scar
      end;

    HtmlCommentScope:
      begin
        if Start.fOffset>3 then backstart:= 3
                           else backstart:= Start.fOffset;
        t:= Start.fPar^.Text+ Start.fOffset-backstart;
        if t<>nil then
          begin
            if Stop.Par^.Length > Stop.ParOffset+3 then backend:= 3
                                                   else backend:= Stop.fPar^.Length - Stop.fOffset;
            searchlen:= Stop.fOffset+backend - (Start.fOffset-backstart);
            scar:= t[searchlen];
            keyfound:= StrPos(t, PChar(EndComment));
            if keyfound<>nil then
              begin
                Result:= True;
                stopendpos:= Start.Pos - backstart + (keyfound-t) + 3;
                Stop.Assign(Start);
                while Stop.ForwardToDyn(stopendpos) do Stop.RemoveDyn;
                Stop.Pos:= stopendpos;
                {FillDyn(Bracket, Stop, False, HtmlStopScope, BaseIndex, 1, 2);
                Stop.Pos:= Stop.Pos+1;}
                dinfo.DynStyle:= 0;
                Stop.AddDyn(dinfo)
              end;
            t[searchlen]:= scar
          end
      end;

    ScriptScope, PerlScriptScope:
      begin
        if curkey-BaseIndex = HtmlScriptIndex then
          begin
            if Start.fOffset>Length(EndHtmlScript) then backstart:= Length(EndHtmlScript)
                                                   else backstart:= Start.fOffset;
            if Stop.Par^.Length > Stop.ParOffset+Length(EndHtmlScript) then backend:= Length(EndHtmlScript)
                                                                       else backend:= Stop.fPar^.Length - Stop.fOffset;
            searchlen:= Stop.fOffset+backend - (Start.fOffset-backstart);
            if searchlen>0 then
              begin
                t:= TPlusMemo(Start.fPMemo).GetUpText(Start.fPar, Start.fParNb, Start.fOffset-backstart, searchlen);
                scar:= t[searchlen];
                keyfound:= StrPos(t, PChar(EndHtmlScript));
                if keyfound<>nil then
                  begin
                    Result:= True;
                    Stop.Pos:= Start.Pos - backstart + (keyfound-t);
                    Stop.RightOfDyn;
                    FillDyn(Bracket, Stop, True, HtmlStopScope, BaseIndex, Length(EndHtmlScript)-1, Length(EndHtmlScript));
                    Stop.Pos:= Stop.Pos+Length(EndHtmlScript)-1;
                    dinfo.DynStyle:= 0;
                    Stop.AddDyn(dinfo)
                  end;
                t[searchlen]:= scar
              end
          end

        else
          begin
            if Start.fOffset>2 then backstart:= 2
                               else backstart:= Start.fOffset;
            t:= Start.fPar^.Text+ Start.fOffset-backstart;
            if Stop.Par^.Length > Stop.ParOffset+2 then backend:= 2
                                                   else backend:= Stop.fPar^.Length - Stop.fOffset;
            searchlen:= Stop.fOffset+backend - (Start.fOffset-backstart);
            if searchlen>0 then
              begin
                scar:= t[searchlen];
                if curkey-BaseIndex = PerlScriptIndex then keyfound:= StrPos(t, PChar(EndPerlScript))
                                                      else keyfound:= StrPos(t, PChar(EndScript));
                if keyfound<>nil then
                  begin
                    Result:= True;
                    Stop.Pos:= Start.Pos - backstart + (keyfound-t) + 1;
                    FillDyn(Bracket, Stop, False, HtmlStopScope, BaseIndex, 1, 2);
                    Stop.Pos:= Stop.Pos+1;
                    dinfo.DynStyle:= 0;
                    Stop.AddDyn(dinfo)
                  end;
                t[searchlen]:= scar
              end
          end
      end;

    SpecialCharScope:
      begin
        t:= Start.fPar^.Text + Start.fOffset;
        searchlen:= Stop.ParOffset - Start.fOffset;
        keyfound:= t;
        if (Start.fOffset=0) or (Start.fPar^.Text[Start.fOffset-1]<>';') then
          while (searchlen>0) and (keyfound^ in ['a'..'z', 'A'..'Z', '0'..'9', '&', '_', '#']) do
            begin
              Inc(keyfound);
              Dec(searchlen)
            end;
        if (searchlen>0) or (Stop.fOffset=Start.fPar^.Length) then
          begin
            Result:= True;
            Stop.Assign(Start);
            Stop.Pos:= Start.Pos + (keyfound - t);
            if (keyfound^=';') then Stop.Pos:= Stop.Pos + 1;
            dinfo.DynStyle:= 0;
            Stop.AddDyn(dinfo)
          end
      end
    end;

end;     { FindStop }

function SpecStrScan(t: PChar): PChar;
  { special scan for either an opening bracket (<) or ampersand
    used for searching a start key in base context }
begin
  while (t^<>#0) and (t^<>'<') and (t^<>'&') do Inc(t);
  if t^=#0 then Result:= nil else Result:= t
end;


function THtmlHighlighter.FindStart(Start, Stop: TPlusNavigator; BaseIndex: Integer): Boolean;

var slen    : PlusCardinal;
    t       : PChar;
    tcomp1, tcomp2: PChar;
    i       : PlusCardinal;
    curscope : Integer;
    scar     : Char;
    startword: PlusCardinal;
    endword  : PlusCardinal;
    wordlen  : PlusCardinal;
    searchlen: PlusCardinal;
    compelem : Boolean;
    kindex   : Integer;
    ahilit   : THighlightInfo;
    actx     : Integer;
    dinfo    : DynInfoRec;

label SubH;

begin
  Start.RightOfDyn;
  Result:= False;
  if getlevel(Start.pDynAttr^)>=15 then Exit;

  with Start.pDynAttr^ do
    if DynStyle and $80 = 0 then curscope:= 0
                            else curscope:= Context;


  case curscope of
    0:  if Stop.ParOffset>Start.fOffset then
          begin
            searchlen:= Stop.fOffset - Start.fOffset;
            t:= Start.fPar^.Text + Start.fOffset;
            scar:= t[searchlen];
            t[searchlen]:= #0;
            tcomp1:= SpecStrScan(t);
            if tcomp1<>nil then
              begin
                Result:= True;
                Stop.Pos:= Start.Pos + (tcomp1-t);
                Stop.RightOfDyn;
                t[searchlen]:= scar;
                if tcomp1^='<' then
                  if (tcomp1[1]='!') and (tcomp1[2]='-') and (tcomp1[3]='-') then
                    begin
                      FillDyn(HtmlComment, Stop, True, HtmlCommentScope, BaseIndex + HtmlCommentIndex, 3, 4);
                      Stop.Pos:= Stop.Pos + 4
                    end
                  else
                    begin
                      FillDyn(Bracket, Stop, True, HtmlStartScope, BaseIndex, 1, 1);
                      Stop.Pos:= Stop.Pos+1
                    end
                else      { Highlighting &nbsp or other like this }
                  begin
                    FillDyn(SpecialChars, Stop, True, SpecialCharScope, BaseIndex + SpecialCharIndex, 1, 1);
                    Stop.Pos:= Stop.Pos+1
                  end
              end { found start of tag }
            else t[searchlen]:= scar
          end;

    HtmlStartScope: { Start is normally positioned right after the opening bracket }
      if Start.fPar^.Text<>nil then
        begin
          t:= Start.fPar^.Text;
          startword:= Start.fOffset;

          if t[startword]='%' then
            begin
              Result:= True;
              FillDyn(ScriptContent, Start, False, ScriptScope, BaseIndex + PcScriptIndex, 1, 2);
              Stop.Assign(Start);
              Stop.Pos:= Stop.Pos + 1;
              goto SubH
            end;

          if t[startword]='?' then
            begin
              Result:= True;
              FillDyn(ScriptContent, Start, False, PerlScriptScope, BaseIndex + PerlScriptIndex, 1, 2);
              Stop.Assign(Start);
              Stop.Pos:= Stop.Pos + 1;
              goto SubH
            end;

          { find the end of word }
          endword:= startword;
          while not (t[endword] in [' ', #0, '>']) do Inc(endword);
          wordlen:= endword-startword;

          if not fXMLSyntax then
            begin
              t:= TPlusMemo(Start.fPMemo).GetUpText(Start.fPar, Start.fParNb, startword, wordlen);

              kindex:= 0;
              if t^='/' then
                begin
                  compelem:= True;
                  Dec(wordlen);
                  Inc(t)
                end
              else compelem:= False;

              for i:= 0 to HtmlKeywordCount-1 do
                begin
                  slen:= Length(HtmlList[i].Element);
                  if slen=wordlen then
                    begin
                      tcomp1:= PChar(HtmlList[i].Element);
                      tcomp2:= t;
                      while (slen>0) and (tcomp1^=tcomp2^) do
                        begin
                          Dec(slen);
                          Inc(tcomp1);
                          Inc(tcomp2)
                        end;

                      if (slen=0) and ((not compelem) or HtmlList[i].Complement) then
                        begin
                          kindex:= i;
                          Result:= True;
                          Break
                        end
                    end
                end
            end
          else
            begin
              Result:= True;
              compelem:= False;
              kindex:= 1
            end;

          if compelem then Inc(wordlen);
          Stop.Assign(Start);
          while Stop.ForwardToDyn(Start.Pos + wordlen) do Stop.RemoveDyn;

          if Result then
            begin
              FillDyn(HtmlKeyword, Start, False, HtmlKeywordScope, BaseIndex + kindex, wordlen, wordlen);
              if compelem or (not (HtmlList[kindex].Attributes)) then
                begin
                  ahilit:= UnknownTag;
                  actx  := HtmlUnknownScope
                end
              else
                begin
                  ahilit:= HtmlTag;
                  actx  := HtmlTagScope
                end;

              Stop.Assign(Start);
              Stop.Pos:= Start.Pos + wordlen;
              FillDyn(ahilit, Stop, False, actx, BaseIndex+kindex, 0, 0)
            end

          else
            begin
              Result:= True;
              FillDyn(UnknownTag, Start, False, HtmlUnknownScope, BaseIndex+HtmlUnknownIndex, wordlen, 0);
              Stop.Assign(Start);
              Stop.Pos:= Start.Pos + wordlen;
            end
        end;   { HtmlStartContext }

    HtmlTagScope:
      if Start.fPar^.Text<>nil then
        begin
          t:= Start.fPar^.Text + Start.fOffset;
          searchlen:= Stop.fOffset-Start.fOffset;
          scar:= t[searchlen];
          t[searchlen]:= #0;
          tcomp1:= HStrScan(t, '=');
          t[searchlen]:= scar;
          if tcomp1<>nil then
            begin
              Result:= True;
              dinfo:= Start.DynAttr;
              startword:= Start.fOffset + (tcomp1-t);
              t:= Start.fPar^.Text;
              endword:= startword+1;
              while (startword>0) and (t[startword-1]=' ') do Dec(startword);
              while (startword>0) and (t[startword-1]<>' ') do Dec(startword);
              Stop.Pos:= Start.fPar^.StartOffset + startword;
              FillDyn(HtmlAttribute, Stop, False, HtmlAttributeScope, BaseIndex, endword-startword, endword-startword);
              Stop.Pos:= Start.fPar^.StartOffset + endword;
              wordlen:= startword;
              startword:= endword;
              while t[endword]=' ' do Inc(endword);
              if t[endword]='"' then
                begin
                  Inc(endword);
                  while not (t[endword] in ['"', #0]) do Inc(endword);
                  if t[endword]='"' then Inc(endword)
                end
              else
                if t[endword]='''' then
                  begin
                    Inc(endword);
                    while not (t[endword] in ['''', #0]) do Inc(endword);
                    if t[endword] = '''' then Inc(endword)
                  end
                else
                  while not (t[endword] in [' ', '>', #0]) do Inc(endword);
              while Stop.ForwardToDyn(Start.fPar^.StartOffset+endword) do
                                                             Stop.RemoveDyn;
              if Start.fOffset>=wordlen then
                begin
                  Start.fDynNb:= High(Start.fDynNb);
                  Start.RightOfDyn
                end;
              Stop.Pos:= Start.fPar^.StartOffset + startword;
              FillDyn(HtmlAttributeValue, Stop, False, HtmlAttributeValueScope, BaseIndex, endword-startword, endword-startword);
              Stop.Pos:= Start.fPar^.StartOffset+endword;
              Stop.AddDyn(dinfo)
            end;
      end
    end;
SubH:
  if SubHighlighter<>nil then Result:= Result or (inherited FindStart(Start, Stop, BaseIndex+HtmlIndexCount))
end;    { FindStart }


end.


