unit pmpprev;

{ Copyright (c) Electro-Concept Mauricie, 1997-1998-1999-2000 }
{ TPlusMemoPrinter accessory component:
    This unit contains the Preview form and display processing

  Dec. 22, 2000 release
    Added #12 as page break if found at the start of a paragraph
    Added PrintRange, PrintSelection and PrintFromDialog methods
    Added possibility to print only a range of pages from the preview window

  Oct. 21, 1999 release
    Added FooterFont, HeaderFont properties
    Header and footers now handle multiline text
    Zoom capability

  Initial release March 8, 1999 }

interface

uses
  WinTypes, WinProcs, Messages, SysUtils, Classes, Graphics, Controls, Forms, Dialogs,
  ExtCtrls, StdCtrls, Buttons, PMPrint, PlusMemo;

type
  TPreviewMouseState = (pmsNormal, pmsLeft, pmsRight, pmsTop, pmsBottom);
  TFrmPMPrintPreview = class(TForm)
    Panel1: TPanel;
    BtnPrint: TButton;
    BtnPrinterSetup: TButton;
    BtnClose: TBitBtn;
    BtnNextPage: TBitBtn;
    BtnPrevious: TBitBtn;
    LblPage: TLabel;
    BtnMargins: TSpeedButton;
    sbPreview: TScrollBox;
    PBPreview: TPaintBox;
    cbZoom: TComboBox;
    procedure sbPreviewResize(Sender: TObject);
    procedure PBPreviewPaint(Sender: TObject);
    procedure FormShow(Sender: TObject);
    procedure FormCreate(Sender: TObject);
    procedure BtnNextPageClick(Sender: TObject);
    procedure BtnPreviousClick(Sender: TObject);
    procedure BtnPrinterSetupClick(Sender: TObject);
    procedure PBPreviewMouseMove(Sender: TObject; Shift: TShiftState; X,
      Y: Integer);
    procedure PBPreviewMouseDown(Sender: TObject; Button: TMouseButton;
      Shift: TShiftState; X, Y: Integer);
    procedure PBPreviewMouseUp(Sender: TObject; Button: TMouseButton;
      Shift: TShiftState; X, Y: Integer);
    procedure BtnPrintClick(Sender: TObject);
    procedure BtnMarginsClick(Sender: TObject);
    procedure FormDestroy(Sender: TObject);
    procedure cbZoomChange(Sender: TObject);
  private
    { Private declarations }

    { position of margins in the PBPreview paint box }
    fLeftMargin, fRightMargin, fTopMargin, fBottomMargin: Integer;

    fMouseState    :  TPreviewMouseState;
    fLastPos       : TPoint;    { last position where a dragging margin has been drawn }
    fCurrentPage   : Integer;
    fBufCanvas     : TCanvas;   { a canvas attached to a printer device context used when
                                  painting the preview lines in the PBPreview paint box }
    fPrintDialog   : TPrintDialog;  { created as needed, when user presses Print Setup button the first time }

    procedure SetLblPage;       { update the label that indicates the current page number }
    procedure setCurrentPage(cp: Integer);
    procedure PaintMargins;     { paint the dashed margin lines on the PBPreview paint box }
    procedure DrawLinesSpecial(fl, ll: Longint; sLeft, sTop, LineSpace, sZoom: Single);

  public
    { Public declarations }
    ZoomFact: Single;           { zoom factor between printer pixels and PBPreview pixels }
    PMPrinter: TPlusMemoPrinter;
    property CurrentPage: Integer read fCurrentPage write setCurrentPage;  { current page in preview }
  end;

var FrmPMPrintPreview: TFrmPMPrintPreview;

implementation

{$R *.DFM}
uses Printers  {$IFNDEF WIN32} , PlusSup  {$ENDIF} ;

procedure ClearFPUExceptions;
  // clears pending FPU exceptions.
  // useful on wrong "invalid floating point operation" exceptions
  asm
    FNCLEX
  end;


procedure TFrmPMPrintPreview.sbPreviewResize(Sender: TObject);
var ph, pw, havail, ppi, nleft: Integer;
begin
  ClearFPUExceptions;
  PMPrinter.FormatPrint;
  { determine the new zoom factor, set the paint box and margins position }
  ph:= PMPrinter.PageHeight;
  pw:= PMPrinter.PageWidth;
  ppi:= PMPrinter.YPPI;
  havail:= sbPreview.ClientHeight;

  try
    case cbZoom.ItemIndex of
      0..5:
          begin
            if PBPreview.Width<sbPreview.ClientWidth-20 then
              nleft:= (sbPreview.ClientWidth-pbPreview.Width) div 2
            else nleft:= 10
          end;
      6:     { fit page width }
          begin
            ZoomFact:= (sbPreview.ClientWidth-20)/pw;
            nleft:= 10
          end;

      else    { fit page }
        if ph/pw>(havail-20)/(sbPreview.ClientWidth-20) then
          begin
            ZoomFact:= (havail-20)/ph;
            nleft:= Round((sbPreview.ClientWidth-(pw*ZoomFact))/2);
          end
        else
          begin
            ZoomFact:= (sbPreview.ClientWidth-20)/pw;
            nleft:= 10;
          end
      end;

    fLeftMargin  := Round(PMPrinter.MarginLeft*ppi*ZoomFact);
    fRightMargin := Round((pw-PMPrinter.MarginRight*ppi)*ZoomFact);
    fTopMargin   := Round(PMPrinter.MarginTop*ppi*ZoomFact);
    fBottomMargin:= Round((ph-PMPrinter.MarginBottom*ppi)*ZoomFact);
    PBPreview.SetBounds(nleft, 10, Round(PMPrinter.PageWidth*ZoomFact), Round(PMPrinter.PageHeight*ZoomFact))

  except
    fLeftMargin  := 0;
    fRightMargin := 0;
    fTopMargin   := 0;
    fBottomMargin:= 0
  end;

end;

procedure TFrmPMPrintPreview.PBPreviewPaint(Sender: TObject);
var clrect: TRect;    { clipping rectangle of PBPreview }
    firstline, lastline, ppi: Integer;
    pdevice, pdriver, pport : PChar;
    pdevmodehand, dc        : THandle;
    pdevmode                : Pointer;
    linespacingpix          : Integer;
    topmarginpix,
    slinespacing            : Single;

begin
  ClearFPUExceptions;
  with PMPrinter do
    begin
      ppi:= YPPI;
      slinespacing:= LineSpacing;
      if sLineSpacing<0 then linespacingpix:= -Round(LineHeight*sLineSpacing)
                        else linespacingpix:=  Round(sLineSpacing*PPI);
      topmarginpix:= MarginTop*ppi;
    end;

  if linespacingpix=0 then linespacingpix:=1;

  with PBPreview, Canvas do
    begin
      clrect:= ClipRect;
      FillRect(clrect);
      if BtnMargins.Down then PaintMargins;
      { determine which lines need be painted }
      firstline:= Trunc((clrect.Top/ZoomFact-topmarginpix)/linespacingpix);
      if firstline<0 then firstline:= 0;
      lastline:= Trunc((clrect.Bottom/ZoomFact-topmarginpix)/LineSpacingPix) + 1;
      if lastline>=PMPrinter.LinesPerPage then lastline:= PMPrinter.LinesPerPage-1;

      { clip the unprintable area }
      ExcludeClipRect(Handle, 0, 0, Trunc(PMPrinter.PageXStart*ZoomFact), ClientHeight);
      ExcludeClipRect(Handle, 0, 0, ClientWidth, Trunc(PMPrinter.PageYStart*ZoomFact));
      ExcludeClipRect(Handle, 0, Trunc((PMPrinter.PrintHeight+PMPrinter.PageYStart)*ZoomFact)+1,
                                                 ClientWidth, ClientHeight);

      ExcludeClipRect(Handle, Trunc((PMPrinter.PrintWidth+PMPrinter.PageXStart)*ZoomFact)+1, 0,
                                              ClientWidth, ClientHeight)
    end;

  { get the printer device context, attach fBufCanvas to it }
  pdevice:= StrAlloc(1000);
  pdriver:= StrAlloc(1000);
  pport  := StrAlloc(1000);
  Printer.GetPrinter(pdevice, pdriver, pport, pdevmodehand);
  pdevmode:= GlobalLock(pdevmodehand);
  fBufCanvas.Handle:= CreateDC(pdriver, pdevice, pport, pdevmode);
  fBufCanvas.Font.PixelsPerInch:= PMPrinter.YPPI;

  DrawLinesSpecial(Longint(PMPrinter.LinesPerPage)*CurrentPage+firstline,
                   Longint(PMPrinter.LinesPerPage)*CurrentPage+lastline,
                   PMPrinter.MarginLeft*PMPrinter.XPPI,
                   topmarginpix + firstline*LineSpacingPix,
                   LineSpacingPix,
                   ZoomFact);

  { finished with the printer device context }
  dc:= fBufCanvas.Handle;
  fBufCanvas.Handle:= 0;
  DeleteDC(DC);
  StrDispose(pdevice);
  StrDispose(pdriver);
  StrDispose(pport);
  GlobalUnlock(pdevmodehand);
end;     { PBPreviewPaint }

procedure TFrmPMPrintPreview.SetLblPage;
begin
  LblPage.Caption:= 'Page ' + IntToStr(CurrentPage+1) + '/' + IntToStr(PMPrinter.PageCount)
end;

procedure TFrmPMPrintPreview.setCurrentPage(cp: Integer);
begin
  if cp>=PMPrinter.PageCount then cp:= PMPrinter.PageCount-1;
  if cp<>fCurrentPage then
    begin
      fCurrentPage:= cp;
      SetLblPage;
      PBPreview.Invalidate;
      BtnNextPage.Enabled:= fCurrentPage<PMPrinter.PageCount-1;
      BtnPrevious.Enabled:= fCurrentPage>0
    end
end;


procedure TFrmPMPrintPreview.FormShow(Sender: TObject);
begin
  PMPrinter.FormatPrint;
  cbZoom.ItemIndex:= 3;
  cbZoomChange(Self);
  SetLblPage
end;



procedure TFrmPMPrintPreview.FormCreate(Sender: TObject);
begin
  fBufCanvas:= TCanvas.Create;
end;

procedure TFrmPMPrintPreview.FormDestroy(Sender: TObject);
begin
  fBufCanvas.Free
end;


procedure TFrmPMPrintPreview.BtnNextPageClick(Sender: TObject);
begin
  CurrentPage:= CurrentPage+1
end;

procedure TFrmPMPrintPreview.BtnPreviousClick(Sender: TObject);
begin
  CurrentPage:= CurrentPage-1
end;

procedure TFrmPMPrintPreview.BtnPrinterSetupClick(Sender: TObject);
var oldpw, oldph, oldyppi, oldxppi: Integer; doprint: Boolean;
begin
  oldpw:= PMPrinter.PageWidth;
  oldph:= PMPrinter.PageHeight;
  oldyppi:= PMPrinter.YPPI;
  oldxppi:= PMPrinter.XPPI;

  if fPrintDialog=nil then
    begin
      fPrintDialog:= TPrintDialog.Create(Self);
      fPrintDialog.Options:= [poPageNums, poWarning];
      fPrintDialog.MinPage:= 1;
      fPrintDialog.MaxPage:= PMPrinter.PageCount
    end;

  doprint:= fPrintDialog.Execute;
  if ((oldpw<>GetDeviceCaps(Printer.Handle, PHYSICALWIDTH)) or
      (oldph<>GetDeviceCaps(Printer.Handle, PHYSICALHEIGHT)) or
      (oldyppi<>GetDeviceCaps(Printer.Handle, LOGPIXELSY)) or
      (oldxppi<>GetDeviceCaps(Printer.Handle, LOGPIXELSX))) then
      begin
        PMPrinter.MemoToPrint:= PMPrinter.MemoToPrint;  { make a reformat }
        cbZoomChange(Self)  { update the display }
      end;
  if doprint then PMPrinter.PrintFromDialog(fPrintDialog)
end;



procedure TFrmPMPrintPreview.PBPreviewMouseMove(Sender: TObject;
  Shift: TShiftState; X, Y: Integer);
begin
  if not BtnMargins.Down then
    begin
      Cursor:= crDefault;
      Exit
    end;

  with PBPreview.Canvas do
    begin
      { prepare to drag a margin }
      Pen.Color:= clGray;
      Pen.Style:= psDot;
      Pen.Mode:= pmXOR;
      if X<0 then X:= 0;
      if X>ClientWidth-1 then X:= ClientWidth-1;
      if Y<0 then Y:= 0;
      if Y>ClientHeight-1 then Y:= ClientHeight-1;

      case fMouseState of
        pmsNormal: begin
                     if (X=fLeftMargin) or (X=fRightMargin) then Cursor:= crHSplit
                     else
                       if (Y=fTopMargin) or (Y=fBottomMargin) then Cursor:= crVSplit
                                                              else Cursor:= crDefault;
                     WinProcs.SetCursor(Screen.Cursors[Cursor])
                   end;

        pmsLeft, pmsRight:
                 begin
                   if fMouseState=pmsLeft then
                     begin
                       if X>fRightMargin-3 then X:= fRightMargin-3
                     end
                   else
                     if X<fLeftMargin+3 then X:= fLeftMargin+3;

                   MoveTo(fLastPos.X, 0);
                   LineTo(fLastPos.X, PBPreview.Height);
                   MoveTo(X, 0);
                   LineTo(X, PBPreview.Height)
                 end;

        pmsTop, pmsBottom:
                 begin
                   if fMouseState=pmsTop then
                     begin
                       if Y>fBottomMargin-3 then Y:= fBottomMargin-3
                     end
                   else
                     if Y<fTopMargin+3 then Y:= fTopMargin+3;
                   MoveTo(0, fLastPos.Y);
                   LineTo(PBPreview.Width, fLastPos.Y);
                   MoveTo(0, Y);
                   LineTo(PBPreview.Width, Y)
                  end;

        end;   { case fMouseState }
      fLastPos.X:= X;
      fLastPos.Y:= Y
    end;    { with PBPreview }
end;  { PBPreviewMouseMove }

procedure TFrmPMPrintPreview.PBPreviewMouseDown(Sender:  TObject;
  Button: TMouseButton; Shift: TShiftState; X, Y: Integer);
begin
  if not BtnMargins.Down then Exit;
  if X=fLeftMargin then fMouseState:= pmsLeft
  else
    if X=fRightMargin then fMouseState:= pmsRight
    else
      if Y=fTopMargin then fMouseState:= pmsTop
      else if Y=fBottomMargin then fMouseState:= pmsBottom;

  with PBPreview.Canvas do
    begin
      Pen.Color:= clGray;
      Pen.Style:= psDot;
      Pen.Mode:= pmXOR;
      fLastPos.X:= X;
      fLastPos.Y:= Y;

      case fMouseState of
        pmsLeft: begin
                   MoveTo(fLeftMargin, 0);
                   LineTo(fLeftMargin, PBPreview.Height)
                 end;
        pmsRight: begin
                    MoveTo(fRightMargin, 0);
                    LineTo(fRightMargin, PBPreview.Height)
                  end;
        pmsTop:  begin
                   MoveTo(0, fTopMargin);
                   LineTo(PBPreview.Width, fTopMargin)
                 end;
        pmsBottom: begin
                     MoveTo(0, fBottomMargin);
                     LineTo(PBPreview.Width, fBottomMargin)
                   end
        end
    end
end;

procedure TFrmPMPrintPreview.PBPreviewMouseUp(Sender: TObject;
  Button: TMouseButton; Shift: TShiftState; X, Y: Integer);
var
  ppi       : Integer;
  repaginate: Boolean;

begin
  if fMouseState<>pmsNormal then
    with PBPreview.Canvas do
      begin
        Pen.Color:= clGray;
        Pen.Style:= psDot;
        Pen.Mode:= pmXOR;
        if X<0 then X:= 0;
        if X>PBPreview.Width-1 then X:= PBPreview.Width-1;
        if Y<0 then Y:= 0;
        if Y>PBPreview.Height-1 then Y:= PBPreview.Height-1;

        case fMouseState of
          pmsLeft, pmsRight:
                   begin
                     MoveTo(fLastPos.X, 0);
                     LineTo(fLastPos.X, PBPreview.Height);
                   end;
          pmsTop, pmsBottom:
                   begin
                     MoveTo(0, fLastPos.Y);
                     LineTo(PBPreview.ClientWidth, fLastPos.Y);
                    end;
          end;

        ppi:= PMPrinter.YPPI;

        repaginate:= False;
        case fMouseState of
          pmsLeft: begin
                     if X>fRightMargin-10 then X:= fRightMargin-10;
                     if X<>fLeftMargin then
                       begin
                         PMPrinter.MarginLeft:= X/(ZoomFact*ppi);
                         repaginate:= True
                       end
                     end;
          pmsRight: begin
                     if X<fLeftMargin+10 then X:= fLeftMargin+10;
                     if X<>fRightMargin then
                       begin
                         PMPrinter.MarginRight:= (PBPreview.ClientWidth-X)/(ZoomFact*ppi);
                         repaginate:= True
                       end
                     end;
          pmsTop: begin
                     if Y>fBottomMargin-10 then Y:= fBottomMargin-10;
                     if Y<>fTopMargin then
                       begin
                         PMPrinter.MarginTop:= Y/(ZoomFact*ppi);
                         repaginate:= True
                       end
                     end;

          pmsBottom: begin
                     if Y<fTopMargin+10 then Y:= fTopMargin+10;
                     if Y<>fBottomMargin then
                       begin
                         PMPrinter.MarginBottom:= (PBPreview.ClientHeight-Y)/(ZoomFact*ppi);
                         repaginate:= True
                       end
                     end
          end;   { case fMouseState, updating margins }

        fMouseState:= pmsNormal;
        if repaginate then
          begin
            cbZoomChange(Self);
            PBPreview.Invalidate;
            SetLblPage
          end
      end;   { with PBPreview.Canvas }

end;    { PBPreviewMouseUp }

procedure TFrmPMPrintPreview.BtnPrintClick(Sender: TObject);
begin
  PMPrinter.Print
end;

procedure TFrmPMPrintPreview.BtnMarginsClick(Sender: TObject);
begin
  PBPreview.Invalidate;
  PBPreviewMouseMove(Self, [], 0, 0)  { update the cursor and mouse state if needed }
end;

procedure TFrmPMPrintPreview.PaintMargins;
begin
  with PBPreview, Canvas do
    begin
      Pen.Color:= clBlack;
      Pen.Width:= 1;
      Pen.Style:= psDash;
      Pen.Mode:= pmXOR;
      MoveTo(fLeftMargin, 0);
      LineTo(fLeftMargin, ClientHeight);

      MoveTo(fRightMargin, 0);
      LineTo(fRightMargin, ClientHeight);

      MoveTo(0, fTopMargin);
      LineTo(ClientWidth, fTopMargin);

      MoveTo(0, fBottomMargin);
      LineTo(ClientWidth, fBottomMargin)
    end
end;

type TProtPlusMemo = class(TPlusMemo);


procedure TFrmPMPrintPreview.DrawLinesSpecial(fl, ll: Longint; sLeft, sTop, LineSpace, sZoom: Single);
  var dc: hdc;
      i: Longint;
      j, b: PlusCardinal;
      t: PChar;
      tlen: PlusCardinal;
      par: ParInfo;
      lin: LineInfo;
      cCanvas: TCanvas;
      sstyle: TExtFontStyles;
      currentattr: TExtFontStyles;
      currentdyn : DynInfoRec;
      curdynnb   : PlusCardinal;
      logfont:  TLogFont;

      smemo: TProtPlusMemo;
      charwidths: array[Char] of Integer;

  procedure XORStyleCode(var Style: TExtFontStyles; Code: Char);
  var element: TExtFontStyle;  isformatchar: Boolean;

    function XORStyle(style: TExtFontStyles; element: TExtFontStyle): TExtFontStyles;
      begin
      if element in style then Result:= style - [element]
                          else Result:= style + [element]
      end;

  begin
    isformatchar:= True;
    element:= fsBold;  { to avoid a warning }
    case Code of
      ctrlUnderline: element:= fsUnderline;
      ctrlItalic   : element:= fsItalic;
      ctrlBold     : element:= fsBold;
      ctrlHighlight: TPlusFontStyle(element):= TPlusFontStyle(fsHighlight);
      ctrlAltFont  : TPlusFontStyle(element):= TPlusFontStyle(fsAltFont)
      else isformatchar:= False
      end;
    if isformatchar then Style:= XORStyle(Style, element)
  end;

  procedure SetupLogfontStyle(var lf: TLogFont; style: TExtFontStyles);
    var f: TFont; bstyle: Byte;  scolor: TColor;
    begin
    with lf do
      begin
      if TPlusFontStyle(fsAltFont) in TPlusFontStyles(style) then f:= smemo.AltFont
                                                             else f:= smemo.Font;
      fBufCanvas.Font.PixelsPerInch:= PMPrinter.YPPI;
      fBufCanvas.Font:= f;
      if f.Height<0 then fBufCanvas.Font.Height:= (f.Height*PMPrinter.YPPI) div smemo.Font.PixelsPerInch;
      bstyle:= Byte(style);
      bstyle:= bstyle and $0f;
      fBufCanvas.Font.Style:= TFontStyles(bstyle);
      GetCharWidth(fBufCanvas.Handle, 0, 255, charwidths);
      scolor:= PBPreview.Canvas.Font.Color;
      PBPreview.Canvas.Font:= f;
      PBPreview.Canvas.Font.Style:= TFontStyles(bstyle);
      PBPreview.Canvas.Font.Color:= scolor;
      if f.Height>0 then
        PBPreview.Canvas.Font.Height:= Round(f.Height*ZoomFact)
      else
        PBPreview.Canvas.Font.Size:= Round(f.Size*ZoomFact*fBufCanvas.Font.PixelsPerInch/
                                           PBPreview.Canvas.Font.PixelsPerInch);
      dc:= PBPreview.Canvas.Handle
      end;
    end;

   procedure setColors;
    var t, b: Longint;
    begin
    t:= smemo.Font.Color;
    b:= -1;  { Transparent }
    if par.Foregnd<>-1 then t:= par.Foregnd;

    if TPlusFontStyle(fsHighlight) in TPlusFontStyles(sstyle) then
      begin
      t:=smemo.fHTColor;
      b:=smemo.fHBColor
      end;

    with currentdyn do
      if (DynStyle and $80 <>0) and ((Foregnd<>0) or (Backgnd<>0)) then
        begin
        if (par.Foregnd=-1) and (Foregnd<>-1) then t:= Foregnd;
        if (par.Backgnd=-1) and (Backgnd<>-1) then b:= Backgnd
        end;

    PBPreview.Canvas.Font.Color:= t;
    PBPreview.Canvas.Brush.Color:= b;
    if b<>-1 then PBPreview.Canvas.Brush.Style:= bsSolid
             else PBPreview.Canvas.Brush.Style:= bsClear;
    dc:= PBPreview.Canvas.Handle
    end;

  procedure SetDC(var offset: PlusCardinal; maxoffset: PlusCardinal);
    var changed: Boolean; ss: TExtFontStyles; sdnb: Integer;
    begin
    changed:= False;
    if smemo.StaticFormat then
      while (offset<maxoffset) and (t[offset]<=#26) and (t[offset] in CtrlCodesSet) do
            begin
            changed:= True;
            XORStyleCode(currentattr, t[offset]);
            Inc(offset)
            end;

    sdnb:= curdynnb;
    while (curdynnb<par.DynCount) and (par.DynBuffer^[curdynnb].DynOffset<=offset) do
      begin
      Inc(curdynnb);
      end;

    if sdnb<>curdynnb then
      begin
      changed:= True;
      currentdyn:= par.DynBuffer^[curdynnb-1];
      end;

    if changed then
          begin
          ss:= sstyle;
          sstyle:= smemo.AttrToExtFontStyles(currentattr, currentdyn.DynStyle);
          if (TPlusFontStyles(ss)*[TPlusFontStyle(fsHighlight)]<>
              TPlusFontStyles(sstyle)*[TPlusFontStyle(fsHighlight)]) or (sdnb<>curdynnb) then SetColors;

          if (TPlusFontStyles(ss)-[TPlusFontStyle(fsHighlight)])<>
             (TPlusFontStyles(sstyle)-[TPlusFontStyle(fsHighlight)]) then
                                                             SetupLogFontStyle(logfont, sstyle);
          end;
    end;


  procedure getwords(var offset: PlusCardinal; maxoffset: PlusCardinal);
    var offlim: PlusCardinal;
    begin
    if curdynnb<par.DynCount then offlim:= par.DynBuffer^[curdynnb].DynOffset
                              else offlim:= High(offlim);

    if maxoffset>offlim then maxoffset:= offlim;
    if smemo.StaticFormat then
      while (offset<maxoffset) and
            ((t[offset]>#26) or (not (t[offset] in (CtrlCodesSet+[#9])))) do Inc(offset)
    else
      while (offset<maxoffset) and (t[offset]<>#9) do Inc(offset)
    end;


var r: TRect; currentpar: Longint;
    tm: TTextMetric;

    runningX,
    sLineHeight, sLineBase: Integer;

    runningerror, breakerror, breakadd,
    newx, xtoreach, rdispx, k, numchar: Integer;
    rline: Longint;
    oldpar: pParInfo;

type IntArray = array[0..maxint div SizeOf(Integer)-1] of Integer;
     pIntArray = ^IntArray;

var pardx: pIntArray;
    leftmarg: Integer;

begin
ClearFPUExceptions;
fl:= PMPrinter.PrintLineToMemoLine(fl);
repeat
  rline:= PMPrinter.PrintLineToMemoLine(ll);
  if rline=High(ll) then Dec(ll);
until (ll<0) or (rline<High(ll));
if ll>=0 then ll:= rline;

smemo:= TProtPlusMemo(PMPrinter.BufferMemo);
cCanvas:= PBPreview.Canvas;
dc:= cCanvas.Handle;
smemo.fMouseNav.LineNumber:= fl;
currentpar:= smemo.fMouseNav.ParNumber;

SetTextAlign(dc, ta_updatecp or ta_baseline);

if fl<smemo.LineCount then
  begin
  with smemo do
    begin
    {set font attributes and starting value in buffer }
    par:= fMouseNav.fNavLines.LLPar^;
    lin:= fMouseNav.fNavLines[fl-par.StartLine]
    end;
    t:= par.Text;
    tlen:= par.Length;
    currentattr:= lin.StartAttrib;
    curdynnb:= 0;
    while (curdynnb<par.DynCount) and (par.DynBuffer^[curdynnb].DynOffset<=lin.Start) do Inc(curdynnb);
    if curdynnb=0 then currentdyn:= par.StartDynAttrib^
                  else currentdyn:= par.DynBuffer^[curdynnb-1];
    sstyle:= smemo.AttrToExtFontStyles(currentattr, currentdyn.DynStyle);
    j:= lin.Start;

    SetUpLogFontStyle(logfont, sstyle);

    GetTextMetrics(dc, tm);

    sLineHeight:= tm.tmHeight+1;
    sLineBase:= sLineHeight - tm.tmDescent-1;
    SetColors;

    leftmarg:= Round(PMPrinter.MarginLeft*PMPrinter.XPPI);
    for i:= fl to ll do
      if i<smemo.LineCount then
        begin
        runningX:= leftmarg;

        r.Left:= Round(runningX*ZoomFact);
        rdispX:= r.Left;
        r.Top:= Round((sTop+(i-fl)*LineSpace)*ZoomFact);
        r.Right:= Round((PMPrinter.PageWidth-PMPrinter.MarginRight*PMPrinter.XPPI));
        r.Bottom:= r.Top+sLineHeight;

        if par.Backgnd<>-1 then
          with PBPreview.Canvas do
            begin
            smemo.Brush.Assign(Brush);    { use smemo.Brush as temporary store for brush attributes }
            Brush.Color:= par.Backgnd;
            Brush.Style:= bsSolid;
            Canvas.FillRect(r);
            Brush.Assign(smemo.Brush)
            end;

        if (not smemo.Justified) or (lin.JustifyStart>=lin.Stop) then
          case smemo.Alignment of
            taRightJustify: r.Left:= Round((r.Right - lin.LineWidth)*ZoomFact);
            taCenter: r.Left:= Round(((leftmarg+r.Right-lin.LineWidth) div 2)*ZoomFact)
            end;
        MoveToEx(dc, r.Left, sLineBase+r.Top, nil);

        j:= lin.Start;
        if smemo.Justified and (lin.Spaces>0) then
          begin
          runningerror:= PMPrinter.PageWidth-Round((PMPrinter.MarginLeft+PMPrinter.MarginRight)*PMPrinter.XPPI)-lin.LineWidth;
          breakadd:= runningerror div lin.Spaces;
          breakerror:= runningerror mod lin.Spaces
          end
        else begin
             breakadd:= 0;
             breakerror:= 0
             end;
        runningerror:= breakerror div 2;

        { draw the text }
        repeat
        setDC(j, lin.Stop);
        if j<lin.Stop then
          begin
          b:= j;
          getWords(j, lin.Stop);

          numchar:= j-b;
          if numchar>0 then
            begin
            GetMem(pardx, numchar*SizeOf(Integer));
            for k:=0 to numchar-1 do
               begin
               newx:= runningX+charwidths[t[b+k]];
               if (t[b+k]=' ') and smemo.Justified then
                 begin
                 Inc(newx, breakadd);
                 Inc(runningerror, breakerror);
                 if runningerror>=lin.Spaces then
                   begin
                   Inc(newx);
                   runningerror:= runningerror-lin.Spaces
                   end
                 end;
               xtoreach:= Round(newx*ZoomFact);
               pardx^[k]:= xtoreach-rdispX;
               rdispX:= xtoreach;
               runningX:= newx
               end;

            ExtTextOut(dc, 0, 0, 0, @r, t+b, j-b, PInteger(pardx));
            FreeMem(pardx, numchar*SizeOf(Integer))
            end;

          if (j<lin.Stop) and (t[j]=#9) then
              begin
              runningX:= leftmarg + ((runningX-leftmarg) div
                                     (smemo.TabStops*charwidths[' '])+1)*(smemo.TabStops*charwidths[' ']);
              rdispX:= Round(runningX*ZoomFact);

              MoveToEx(dc, rdispX, sLineBase+r.Top, nil);
              Inc(j);
              end;
          end

        until (j>=lin.Stop) or (j>= tlen);


        { finished for this line, go on with next one }
        if (i>=par.StartLine+par.NbLines-1) and (i<smemo.LineCount-1) then
            begin
            if curdynnb<par.DynCount then SetDC(j, lin.Stop);
            Inc(currentpar);
            oldpar:= smemo.fMouseNav.fNavLines.LLPar;
            smemo.fMouseNav.fNavLines.LLPar:= smemo.fParagraphs.Pointers[currentpar];
            par:= smemo.fMouseNav.fNavLines.LLPar^;
            if oldpar^.Foregnd<>par.Foregnd then SetColors;

            t:= par.Text;
            tlen:= par.Length;
            curdynnb:= 0;
            end;
        if (i<smemo.LineCount-1) and (i<ll) then
          lin:= smemo.fMouseNav.fNavLines.Items[i+1-par.StartLine]
        end;

    smemo.fMouseNav.fNavLines.LLPar:= smemo.fMouseNav.fPar
    end;

end;      { DrawLinesSpecial }


procedure TFrmPMPrintPreview.cbZoomChange(Sender: TObject);
begin
  ClearFPUExceptions;
  sbPreview.HorzScrollBar.Position:= 0;
  sbPreview.VertScrollBar.Position:= 0;
  ZoomFact:= Screen.PixelsPerInch/PMPrinter.YPPI;
  case cbZoom.ItemIndex of
    0: ZoomFact:= 5*ZoomFact;
    1: ZoomFact:= 2*ZoomFact;
    2: ZoomFact:= 1.5*ZoomFact;
    4: ZoomFact:= 0.75*ZoomFact;
    5: ZoomFact:= 0.5*ZoomFact;
    {6, 7: ZoomFact will be set in resizing the scroll box }
    end;
  sbPreviewResize(cbZoom)
end;

end.
